from fastapi import FastAPI, Body
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder
import os
import signal
from rockai.predictor import BasePredictor
import uvicorn
from rockai.parser.config_util import (
    parse_config_file,
    get_predictor_class_name,
    get_predictor_path,
)
from rockai.server.utils import (
    load_class_from_file,
    get_input_type,
    get_output_type,
)
from starlette.responses import StreamingResponse
import rockai.data_class
import typing
import logging
from rockai.data_class import PredictionResponse
from pathlib import Path
from fastapi import Path as FastApiPath
from typing import Any
from rockai.server.runner import Runner, RunnerResult
import uuid
from .json import upload_files
from .files import upload_file
from datetime import datetime
from enum import Enum, auto, unique

# Set up logging
logging.basicConfig(level=logging.DEBUG)  # Set the initial logging level to INFO

# Create a logger
logger = logging.getLogger(__name__)


@unique
class Health(Enum):
    UNKNOWN = auto()
    STARTING = auto()
    READY = auto()
    BUSY = auto()
    SETUP_FAILED = auto()


class PredictorState:
    health: Health


class MyFastAPI(FastAPI):
    predictor_state: PredictorState


def load_predictor_from_file(path) -> BasePredictor:
    pred: BasePredictor = load_class_from_file(
        Path.cwd() / get_predictor_path(parse_config_file(path / "rock.yaml")),
        get_predictor_class_name(parse_config_file(path / "rock.yaml")),
        BasePredictor,
    )
    return pred


def create_app(path: Path) -> MyFastAPI:

    app: MyFastAPI = MyFastAPI()
    app.predictor_state = PredictorState()
    app.predictor_state.health = Health.UNKNOWN
    pred: BasePredictor = load_predictor_from_file(path)

    input_type = get_input_type(pred)

    output_type = get_output_type(pred)

    runner = Runner()

    class PredictionRequest(
        rockai.data_class.PredictionRequest.get_pydantic_model(input_type=input_type)
    ):
        pass

    InfereceResult = PredictionResponse.get_pydantic_model(
        input_type=input_type, output_type=output_type
    )

    @app.on_event("startup")
    async def start_up_event():
        """
        Run the setup function of the predictor and load the model
        """
        logger.debug("setup start...")
        app.predictor_state.health = Health.STARTING
        try:
            pred.setup()
            app.predictor_state.health = Health.READY
            logger.debug("setup finished")
        except Exception as e:
            app.predictor_state.health = Health.SETUP_FAILED
            logger.error(f"Error setting up predictor: {str(e)}")

    @app.post(
        "/predictions",
        response_model=InfereceResult,
        response_model_exclude_unset=True,
    )
    def predict(
        request_body: PredictionRequest = Body(default=None),
    ) -> typing.Any:
        """
        Running the prediction.
        """
        logger.debug("prediction called...")
        if request_body is None or request_body.input is None:
            request_body = PredictionRequest(input={})
        request_body = request_body.dict()
        id = uuid.uuid4().hex
        start_time = datetime.now()
        # result = await runner.predict(id, pred.predict, request_body)
        result = pred.predict(**request_body["input"])
        end_time = datetime.now()
        time_taken = (end_time - start_time).total_seconds()
        final_result = upload_files(
            result, upload_file=lambda fh: upload_file(fh, None)  # type: ignore
        )
        return JSONResponse(
            content=jsonable_encoder(
                InfereceResult(
                    input=request_body["input"],
                    inference_time=time_taken,
                    output=final_result,
                    id=id,
                    started_at=start_time,
                    completed_at=end_time,
                )
            )
        )

    @app.post(
        "/predictions/{prediction_id}",
        response_model=InfereceResult,
        response_model_exclude_unset=True,
    )
    def predic_with_id(
        prediction_id: str = FastApiPath(title="prediction ID"),
        request_body: PredictionRequest = Body(default=None),
    ) -> typing.Any:
        """
        Running the prediction.
        """
        logger.debug("prediction called... ID -> {}".format(prediction_id))
        if request_body is None or request_body.input is None:
            request_body = PredictionRequest(input={})
        request_body = request_body.dict()
        start_time = datetime.now()
        # result = await runner.get_result(pred,request_body)
        result = pred.predict(**request_body["input"])
        # result = await runner.predict(prediction_id, pred.predict, request_body)
        end_time = datetime.now()
        time_taken = (end_time - start_time).total_seconds()
        final_result = upload_files(
            result, upload_file=lambda fh: upload_file(fh, None)
        )
        logger.debug(result)
        if isinstance(result, RunnerResult):
            return JSONResponse(
                content=jsonable_encoder({"msg": result.msg}), status_code=400
            )
        return JSONResponse(
            content=jsonable_encoder(
                InfereceResult(
                    input=request_body["input"],
                    inference_time=time_taken,
                    output=final_result,
                    id=prediction_id,
                    started_at=start_time,
                    completed_at=end_time,
                )
            ),
            status_code=200,
        )

    @app.post("/stream/predictions", response_model_exclude_unset=True)
    def stream_predict(
        request_body: PredictionRequest = Body(default=None),
    ) -> typing.Any:
        """
        Running the prediction.
        """
        logger.debug("streamprediction called...")
        if request_body is None or request_body.input is None:
            request_body = PredictionRequest(input={})
        request_body = request_body.dict()

        def stream_fun():
            for data in pred.predict(**request_body["input"]):
                yield f"event: output\ndata: {data}\n\n"
            yield "event: done\ndata: {}\n\n".format({})

        return StreamingResponse(stream_fun(), media_type="text/event-stream")

    @app.post("/predictions/{prediction_id}/cancel")
    async def cancel(prediction_id: str) -> Any:
        result = runner.cancel(prediction_id)
        """Cancel prediction by id"""
        logger.debug("cancel prediction start...{}".format(prediction_id))
        return JSONResponse(
            content={
                "message": "Prediction {} is cancelled -> {}".format(
                    prediction_id, result
                ),
                "is_canceled": result,
                "prediction_id": prediction_id,
            },
            status_code=200,
        )

    @app.post("/shutdown")
    async def shutdown():
        """
        Shutdown the server.
        """
        runner.clean_up()
        pid = os.getpid()
        os.kill(pid, signal.SIGINT)
        return JSONResponse(content={"message": "Shutting down"}, status_code=200)

    @app.get("/")
    async def root():
        """
        Hello World!, when you see this message, it means the server is up and running.
        """
        return JSONResponse(
            content={"docs_url": "/docs", "model_schema": "/openapi.json"},
            status_code=200,
        )

    @app.get("/health-check")  # type: ignore[return-value]
    async def health_check() -> Any:
        return JSONResponse(content={"status": str(app.predictor_state.health.name)})

    # finally create the application
    return app


def start_server(port):
    app = create_app(path=Path.cwd())
    uvicorn.run(app, host="0.0.0.0", port=port)



from pydantic import BaseModel, create_model,ConfigDict
import typing as t
from rockai.server.types import URLPath
from pathlib import Path
from datetime import datetime
from typing import Optional, Dict, Any


class InferenceBase(BaseModel, extra="allow"):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    input: Dict[str, Any]


class PredictionRequest(InferenceBase):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: Optional[str] = None
    created_at: Optional[datetime] = None

    @classmethod
    def get_pydantic_model(cls, input_type: t.Type[t.Any]) -> t.Any:
        # dynamic_model = create_model(cls.__name__, __base__=cls, input=(input_type,...))
        dynamic_model = create_model(
            cls.__name__, __base__=cls, input=(input_type, None)
        )
        return dynamic_model


class PredictionResponse(InferenceBase):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    output: t.Any
    id: t.Optional[str] = None
    started_at: t.Optional[datetime] = None
    completed_at: t.Optional[datetime] = None
    inference_time: float = 0
    logs: str = None
    error: t.Optional[str] = None
    metrics: t.Optional[t.Dict[str, t.Any]] = None

    @classmethod
    def get_pydantic_model(
        cls, input_type: t.Type[t.Any], output_type: t.Type[t.Any]
    ) -> t.Any:
        return create_model(
            cls.__name__,
            __base__=cls,
            input=(t.Optional[input_type], None),
            output=(output_type, None),
        )


class BaseInput(BaseModel):

    class Config:
        # When using `choices`, the type is converted into an enum to validate
        # But, after validation, we want to pass the actual value to predict(), not the enum object
        use_enum_values = True

    def cleanup(self) -> None:
        """
        Cleanup any temporary files created by the input.
        Later date to added file remover.
        """
        for _, value in self:
            # Handle URLPath objects specially for cleanup.
            if isinstance(value, URLPath):
                value.unlink()
            # Note this is pathlib.Path, which cog.Path is a subclass of. A pathlib.Path object shouldn't make its way here,
            # but both have an unlink() method, so may as well be safe.
            elif isinstance(value, Path):
                try:
                    value.unlink()
                except FileNotFoundError:
                    pass



import json
import requests
from pathlib import Path
import logging
from rich.progress import Progress, SpinnerColumn, TextColumn
import subprocess
import uuid
from rockai.docker.torch_version import torch_version_list
import importlib.metadata

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("docker").setLevel(logging.WARNING)


def remove_some_libs(file_name, lib_to_be_deleted):
    try:
        # Open the file in read mode
        with open(file_name, "r") as file:
            lines = file.readlines()

        # Open the file in write mode
        with open(file_name, "w") as file:
            for line in lines:
                if lib_to_be_deleted not in line:
                    file.write(line)

        print("Successfully removed {} from {}".format(lib_to_be_deleted, file_name))
    except FileNotFoundError:
        print("{} not found".format(file_name))


def get_tensorflow_version(filename: str) -> str:
    """
    Reads a pip's requirements.txt file and returns the TensorFlow version specified in it.
    """
    try:
        with open(filename, "r") as file:
            for line in file:
                if line.startswith("tensorflow=="):
                    # Split the line at '==' to get the package and version
                    parts = line.strip().split("==")
                    if len(parts) == 2:
                        return parts[1]
                    else:
                        return "No specific version specified"
        return None
    except FileNotFoundError:
        return "File not found"


def run_command_and_get_output():
    # Run the command and capture the output
    result = subprocess.run(
        ["python", "-m", "rockai.command.openapi_schema"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    # Check if the command was successful
    if result.returncode == 0:
        # Store the output in a string
        output = result.stdout
    else:
        # Handle the error
        output = f"Error: {result.stderr}"

    return "'{}'".format(
        str(json.dumps(json.dumps(json.loads(output)), separators=(",", ":")))
    )


def build_docker_image(
    image_tag,
    config_map,
    tag,
    platform="linux/amd64",
    port=8000,
    is_tensorflow=False,
    install_cmd=None,
):
    # sed -i s@/archive.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list
    # sed -i s@/security.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list
    # apt-get clean
    # apt-get update

    docker_list = []
    docker_list.append(add_base(image_tag))
    docker_list.append(add_env("DEBIAN_FRONTEND=noninteractive"))
    docker_list.append(add_work_dir("/src"))
    docker_list.append(copy_files(".", "/src"))
    # docker_list.append(copy_files("rock.yaml", "/src"))
    # docker_list.append(
    #     copy_files("{}".format(config_map["predict"].split(":")[0]), "/src")
    # )
    docker_list.append(
        add_run(
            "sed -i s@/archive.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list"
        )
    )
    docker_list.append(
        add_run(
            "sed -i s@/security.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list"
        )
    )
    docker_list.append(add_run("apt-get clean"))
    docker_list.append(add_run("apt update && apt-get update"))
    docker_list.append(add_env('PATH="/root/.pyenv/shims:/root/.pyenv/bin:$PATH"'))
    docker_list.append(
        add_run(
            """--mount=type=cache,target=/var/cache/apt apt-get update -qq && apt-get install -qqy --no-install-recommends \
	make \
	build-essential \
	libssl-dev \
	zlib1g-dev \
	libbz2-dev \
	libreadline-dev \
	libsqlite3-dev \
	wget \
	curl \
	llvm \
	libncurses5-dev \
	libncursesw5-dev \
	xz-utils \
	tk-dev \
	libffi-dev \
	liblzma-dev \
	git \
	ca-certificates \
	&& rm -rf /var/lib/apt/lists/*
"""
        )
    )
    docker_list.append(
        add_run(
            """curl -s -S -L https://raw.githubusercontent.com/pyenv/pyenv-installer/master/bin/pyenv-installer | bash && \
	git clone https://github.com/momo-lab/pyenv-install-latest.git "$(pyenv root)"/plugins/pyenv-install-latest && \
	pyenv install-latest {} && \
	pyenv global $(pyenv install-latest --print {}) && \
	pip install "wheel<1"
""".format(
                config_map["build"]["python_version"],
                config_map["build"]["python_version"],
            )
        )
    )
    docker_list.append(add_expose(port))
    if "python_requirements" not in config_map["build"]:
        logging.info("No python requirements.txt file found in rock.yaml")
    else:
        docker_list.append(
            copy_files("{}".format(config_map["build"]["python_requirements"]), "/src")
        )
        if is_tensorflow:  # building tensorflow
            remove_some_libs(
                config_map["build"]["python_requirements"], "tensorflow-metal"
            )
            remove_some_libs(
                config_map["build"]["python_requirements"], "tensorflow-macos"
            )
            docker_list.append(
                add_run(
                    "pip install -r {} {}".format(
                        "requirements.txt",
                        "-i https://pypi.tuna.tsinghua.edu.cn/simple --extra-index-url https://pypi.org/simple/",
                    )
                )
            )

            tf_version = get_tensorflow_version(
                config_map["build"]["python_requirements"]
            )
            docker_list.append(
                add_run(
                    "pip install tensorflow[and-cuda]=={} -i https://pypi.tuna.tsinghua.edu.cn/simple".format(
                        tf_version
                    )
                )
            )
        else:  # building torch
            remove_some_libs(config_map["build"]["python_requirements"], "torch")
            remove_some_libs(config_map["build"]["python_requirements"], "torchvision")
            remove_some_libs(config_map["build"]["python_requirements"], "torchaudio")
            if install_cmd is not None:
                docker_list.append(add_run(install_cmd))
                docker_list.append(
                    add_run(
                        "pip install -r {} {}".format(
                            "requirements.txt",
                            "-i https://pypi.tuna.tsinghua.edu.cn/simple --extra-index-url https://pypi.org/simple/",
                        )
                    )
                )

    docker_list.append(
        add_run(f"pip install rockai=={importlib.metadata.version('rockai')}")
    )
    docker_list.append(
        add_labels("run.cog.openapi_schema", run_command_and_get_output())
    )
    docker_list.append(add_cmd("rockai start"))

    save_docker_file(docker_list)
    subprocess.run(
        [
            "docker",
            "build",
            "--platform",
            platform,
            "-t",
            tag,
            "-f",
            Path.cwd() / "temp/Dockerfile",
            Path.cwd(),
        ]
    )


def build_cpu_image(
    image_tag, config_map, tag, platform="linux/amd64", port=8000, framework=None
):
    # sed -i s@/archive.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list
    # sed -i s@/security.ubuntu.com/@/mirrors.aliyun.com/@g /etc/apt/sources.list
    # apt-get clean
    # apt-get update

    docker_list = []
    docker_list.append(add_base(image_tag))
    docker_list.append(add_env("DEBIAN_FRONTEND=noninteractive"))
    docker_list.append(add_run("apt update && apt-get update"))
    docker_list.append(add_work_dir("/src"))
    docker_list.append(copy_files(".", "/src"))
    # docker_list.append(copy_files("rock.yaml", "/src"))
    if "python_requirements" in config_map["build"]:

        if framework is not None:
            if framework == "tensorflow":  # Build TF
                tf_version = get_tensorflow_version(
                    config_map["build"]["python_requirements"]
                )
                remove_some_libs(
                    config_map["build"]["python_requirements"], "tensorflow-metal"
                )
                remove_some_libs(
                    config_map["build"]["python_requirements"], "tensorflow-macos"
                )
                docker_list.append(
                    copy_files(
                        "{}".format(config_map["build"]["python_requirements"]), "/src"
                    )
                )

                docker_list.append(
                    add_run(
                        "pip install -r {} {}".format(
                            "requirements.txt",
                            "-i https://pypi.tuna.tsinghua.edu.cn/simple --extra-index-url https://pypi.org/simple/",
                        )
                    )
                )
                docker_list.append(
                    add_run(
                        "pip install tensorflow=={} -i https://pypi.tuna.tsinghua.edu.cn/simple".format(
                            tf_version
                        )
                    )
                )

            else:  # Build Torch
                docker_list.append(
                    copy_files(
                        "{}".format(config_map["build"]["python_requirements"]), "/src"
                    )
                )
                docker_list.append(
                    add_run(
                        "pip install -r {} {}".format(
                            "requirements.txt",
                            "-i https://pypi.tuna.tsinghua.edu.cn/simple --extra-index-url https://download.pytorch.org/whl/cpu",
                        )
                    )
                )

    else:
        logging.info("No python requirements.txt file found in rock.yaml")

    # docker_list.append(
    #     copy_files("{}".format(config_map["predict"].split(":")[0]), "/src")
    # )

    docker_list.append(add_expose(port))

    docker_list.append(
        add_run(f"pip install rockai=={importlib.metadata.version('rockai')}")
    )
    docker_list.append(
        add_labels("run.cog.openapi_schema", run_command_and_get_output())
    )
    docker_list.append(add_cmd("rockai start"))
    save_docker_file(docker_list)
    subprocess.run(
        [
            "docker",
            "build",
            "--platform",
            platform,
            "-t",
            tag,
            "-f",
            Path.cwd() / "temp/Dockerfile",
            Path.cwd(),
        ]
    )


def tf_compat_matrix():
    json_data = [
        {
            "tf_version": "tensorflow-2.16.1",
            "python": "3.9-3.11",
            "cuDNN": "8.9",
            "CUDA": "12.3",
        },
        {
            "tf_version": "tensorflow-2.15.0",
            "python": "3.9-3.11",
            "cuDNN": "8.9",
            "CUDA": "12.2",
        },
        {
            "tf_version": "tensorflow-2.14.0",
            "python": "3.9-3.11",
            "cuDNN": "8.7",
            "CUDA": "11.8",
        },
        {
            "tf_version": "tensorflow-2.13.0",
            "python": "3.8-3.11",
            "cuDNN": "8.6",
            "CUDA": "11.8",
        },
        {
            "tf_version": "tensorflow-2.12.0",
            "python": "3.8-3.11",
            "cuDNN": "8.6",
            "CUDA": "11.8",
        },
        {
            "tf_version": "tensorflow-2.11.0",
            "python": "3.7-3.10",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.10.0",
            "python": "3.7-3.10",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.9.0",
            "python": "3.7-3.10",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.8.0",
            "python": "3.7-3.10",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.7.0",
            "python": "3.7-3.9",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.6.0",
            "python": "3.6-3.9",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.5.0",
            "python": "3.6-3.9",
            "cuDNN": "8.1",
            "CUDA": "11.2",
        },
        {
            "tf_version": "tensorflow-2.4.0",
            "python": "3.6-3.8",
            "cuDNN": "8.0",
            "CUDA": "11.0",
        },
        {
            "tf_version": "tensorflow-2.3.0",
            "python": "3.5-3.8",
            "cuDNN": "7.6",
            "CUDA": "10.1",
        },
        {
            "tf_version": "tensorflow-2.2.0",
            "python": "3.5-3.8",
            "cuDNN": "7.6",
            "CUDA": "10.1",
        },
    ]
    return json_data


def find_correct_image_by_torch_gpu_only(torch_version):
    if torch_version.startswith("1.") and "+" in torch_version:
        parts = torch_version.split("+")
        torch_version = parts[0]

    for each_version in torch_version_list:
        if torch_version == each_version["version"]:
            return each_version["image"], each_version["install"]

    raise Exception(
        "Rock can't find a suppported Cuda or cuDNN version for torch version: {}".format(
            torch_version
        )
    )


def parse_data(result):
    r_list = result["results"]
    result = []
    for item in r_list:
        if (
            "ubuntu" in item["name"]
            and "devel" in item["name"]
            and "cudnn" in item["name"]
        ):
            result.append(item["name"])
    return result


def get_image_tag_from_docker():

    result = []
    page = 1
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(
            description="Downloading image list...(should take around 10-15 seconds)",
            total=None,
        )
        while True:
            url = "https://hub.docker.com/v2/namespaces/nvidia/repositories/cuda/tags?page={}&page_size=100".format(
                page
            )
            response = requests.request("GET", url)

            if response.status_code == 200:
                result += parse_data(response.json())
            elif response.status_code == 404:
                break
            else:
                raise Exception()
            page += 1
        with open("nvidia_docker_cuda_image.json", "w") as file:
            json.dump(result, file)
        return result


def process_requirements(filename):
    with open(filename, "r") as file:
        for line in file:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            line_spl = line.split("==")
            if "tensorflow" == line_spl[0]:
                return "tensorflow", line_spl[1]
            if "torch" == line_spl[0]:
                return "torch", line_spl[1]
    return None, None


def build_final_image(config_map, port=8000):

    if "gpu" in config_map["build"] and config_map["build"]["gpu"] is True:
        if "python_requirements" in config_map["build"]:

            # Build GPU image
            framework, version = process_requirements(
                Path.cwd() / config_map["build"]["python_requirements"]
            )
            if framework == "tensorflow":
                print("Building an TensorFlow image")
                image_tag = "nvidia/cuda:11.8.0-cudnn8-devel-ubuntu22.04"
                build_docker_image(
                    image_tag,
                    config_map,
                    config_map["image"],
                    port=port,
                    install_cmd=None,
                    is_tensorflow=True,
                )
            elif framework == "torch":
                print("Building an PyTorch image {}".format(version))
                image_tag, install_cmd = find_correct_image_by_torch_gpu_only(version)
                build_docker_image(
                    image_tag,
                    config_map,
                    config_map["image"],
                    port=port,
                    install_cmd=install_cmd,
                    is_tensorflow=False,
                )
        else:
            image_tag = "nvidia/cuda:12.5.0-devel-ubuntu22.04"
            build_docker_image(
                image_tag,
                config_map,
                config_map["image"],
                port=port,
                install_cmd=None,
                is_tensorflow=False,
            )
    else:
        py_version = config_map["build"]["python_version"]
        # build CPU image
        if "python_requirements" in config_map["build"]:
            framework, version = process_requirements(
                Path.cwd() / config_map["build"]["python_requirements"]
            )

            py_version = config_map["build"]["python_version"]
            if framework == "tensorflow":
                build_cpu_image(
                    "python:{}".format(py_version),
                    config_map,
                    config_map["image"],
                    port=port,
                    framework="tensorflow",
                )
            elif framework == "torch":
                build_cpu_image(
                    "python:{}".format(py_version),
                    config_map,
                    config_map["image"],
                    port=port,
                    framework="torch",
                )
            else:
                build_cpu_image(
                    "python:{}".format(py_version),
                    config_map,
                    config_map["image"],
                    port=port,
                    framework=None,
                )
        else:
            build_cpu_image(
                "python:{}".format(py_version),
                config_map,
                config_map["image"],
                port=port,
                framework=None,
            )


def add_base(base):
    return "FROM {}\n".format(base)


def add_cmd(cmd_list):
    return "CMD {}\n".format(cmd_list)


def add_expose(port):
    return "EXPOSE {}\n".format(port)


def add_work_dir(dir):
    return "WORKDIR {}\n".format(dir)


def add_run(cmd):
    return "RUN {}\n".format(cmd)


def copy_files(src, dest):
    return "COPY {} {}\n".format(src, dest)


def add_env(env):
    return "ENV {}\n".format(env)


def add_labels(key, value):
    return "LABEL {}={}\n".format(key, value)


def save_docker_file(cmd_list):
    result = "".join(cmd_list)
    # Define the directory path
    directory_path = Path(str(Path.cwd() / "temp"))

    # Check if the directory exists
    if not directory_path.exists():
        # If it does not exist, create the directory
        directory_path.mkdir(parents=True, exist_ok=True)
        print(f"Directory created: {directory_path}")
    else:
        print(f"Directory already exists: {directory_path}")
    try:
        suffix = uuid.uuid4().hex
        with open(Path.cwd() / "temp/Dockerfile".format(suffix), "w") as file:
            file.write(result)
    except Exception as e:
        print("An error occurred while writing to the file. Error: ", str(e))



import yaml

def parse_config_file(file_path):
    with open(file_path, 'r') as file:
        config = yaml.safe_load(file)
    return config


# Example usage
# config_file_path = '/Users/jyz/Desktop/rockai-cli-app/rockai_cli_app/server/test/test_config.yaml'
# parsed_config = parse_config_file(config_file_path)
# print(parsed_config)
def get_predictor_path(config_map):
    predict_config = config_map['predict']
    splited = predict_config.split(":")
    # the first element should a python file, the second element should be Class name
    return splited[0]

def get_predictor_class_name(config_map):
    predict_config = config_map['predict']
    splited = predict_config.split(":")
    # the first element should a python file, the second element should be Class name
    return splited[1]



import base64
import io
import mimetypes
import os
from typing import Optional
from urllib.parse import urlparse
import requests


def upload_file(fh: io.IOBase, output_file_prefix: str = None) -> str:
    fh.seek(0)

    if output_file_prefix is not None:
        name = getattr(fh, "name", "output")
        url = output_file_prefix + os.path.basename(name)
        resp = requests.put(url, files={"file": fh})
        resp.raise_for_status()
        return url

    b = fh.read()
    # The file handle is strings, not bytes
    if isinstance(b, str):
        b = b.encode("utf-8")
    encoded_body = base64.b64encode(b)
    if getattr(fh, "name", None):
        # despite doing a getattr check here, pyright complains that io.IOBase has no attribute name
        # TODO: switch to typing.IO[]?
        mime_type = mimetypes.guess_type(fh.name)[0]  # type: ignore
    else:
        mime_type = "application/octet-stream"
    s = encoded_body.decode("utf-8")
    return f"data:{mime_type};base64,{s}"


def guess_filename(obj: io.IOBase) -> str:
    """Tries to guess the filename of the given object."""
    name = getattr(obj, "name", "file")
    return os.path.basename(name)


def put_file_to_signed_endpoint(
    fh: io.IOBase, endpoint: str, client: requests.Session, prediction_id: Optional[str]
) -> str:
    fh.seek(0)

    filename = guess_filename(fh)
    content_type, _ = mimetypes.guess_type(filename)

    # set connect timeout to slightly more than a multiple of 3 to avoid
    # aligning perfectly with TCP retransmission timer
    connect_timeout = 10
    read_timeout = 15

    headers = {
        "Content-Type": content_type,
    }
    if prediction_id is not None:
        headers["X-Prediction-ID"] = prediction_id

    resp = client.put(
        ensure_trailing_slash(endpoint) + filename,
        fh,  # type: ignore
        headers=headers,
        timeout=(connect_timeout, read_timeout),
    )
    resp.raise_for_status()

    # Try to extract the final asset URL from the `Location` header
    # otherwise fallback to the URL of the final request.
    final_url = resp.url
    if "location" in resp.headers:
        final_url = resp.headers.get("location")

    # strip any signing gubbins from the URL
    return str(urlparse(final_url)._replace(query="").geturl())


def ensure_trailing_slash(url: str) -> str:
    """
    Adds a trailing slash to `url` if not already present, and then returns it.
    """
    if url.endswith("/"):
        return url
    else:
        return url + "/"



import typer
from rockai.server.http import start_server
from rockai.server.utils import is_valid_name
from pathlib import Path
from rockai.parser.config_util import parse_config_file
from typing_extensions import Annotated
from rockai.docker.docker_util import build_final_image
import os
import shutil
import requests
import subprocess
import logging

logger = logging.getLogger()
logging.basicConfig(level=logging.DEBUG)

app = typer.Typer()

APP_NAME = "rockai"


@app.callback()
def callback():
    """
    Callback for the CLI app before any command
    """
    typer.echo("Callback excuted")


@app.command()
def init():
    file = open("rock.yaml", "w")
    file.close()
    file = open("predict.py", "w")
    file.close()


@app.command(name="build")
def build():
    """
    Build the image
    """
    if not os.path.exists(Path.cwd() / "temp"):
        os.makedirs(Path.cwd() / "temp")
        print(f"Folder '{Path.cwd() / 'temp'}' created.")
    else:
        print(f"Folder '{Path.cwd() / 'temp'}' already exists.")

    config_path: Path = Path.cwd() / "rock.yaml"
    if not config_path.is_file():
        raise Exception("rock.yaml config file doesn't exist in the current directory")
    else:
        print("rock.yaml config file exist")

    config_map = parse_config_file(config_path)
    logger.debug(config_map)
    if not is_valid_name(config_map["image"].split("/")[-1]):
        print("Invalid model name, please rename your model accordingly to the following rules")
        print("1. contain no more than 253 characters\n2. contain only lowercase alphanumeric characters,and '-'\n3. start with an alphanumeric character\n4. end with an alphanumeric character")
        raise Exception(
            "Invalid model name: {}".format(config_map["image"].split("/")[-1])
        )
    try:
        # Copy the content of file_1 to file_2
        if "python_requirements" in config_map["build"]:
            shutil.copyfile(
                Path.cwd() / config_map["build"]["python_requirements"],
                Path.cwd() / "temp" / "requirements.txt",
            )
            config_map["build"]["python_requirements"] = "temp/requirements.txt"

    except FileNotFoundError as e:
        raise FileNotFoundError("Source file not found") from e
    except Exception as e:
        raise Exception(f"An error occurred: {e}") from e

    build_final_image(config_map=config_map, port=8000)


@app.command()
def start():
    """
    start local development server
    """
    start_server(8000)


@app.command("push")
def push_model():
    """
    Push the model to the RockAI platform
    """
    build()
    config_path: Path = Path.cwd() / "rock.yaml"
    if not config_path.is_file():
        raise Exception("rock.yaml config file doesn't exist in the current directory")
    else:
        print("rock.yaml config file exist")

    config_map = parse_config_file(config_path)
    subprocess.run(["docker", "image", "push", "{}".format(config_map["image"])])


@app.command(name="login")
def login_to_docker(api_token: Annotated[str, typer.Argument(help="Your API token")],debug: Annotated[bool, typer.Option(help="enable debug mode")] = False):
    url = "https://api.rockai.online/v1/user/docker_token"
    if debug:
        print("Debug mode is enabled")
        url = "https://api-dev.rockai.online:9090/v1/user/docker_token"
    headers = {"Authorization": "Bearer {}".format(api_token)}

    response = requests.get(url, headers=headers)
    response.raise_for_status()
    subprocess.run(
        [
            "docker",
            "login",
            "r.18h.online",
            "-u",
            response.json()["data"]["docker_robot_account"],
            "-p",
            response.json()["data"]["docker_token"],
        ]
    )



import asyncio
from rockai import Client

version = [{'torch_version':'2.3.0','linux':{}}]

# Run a image generation model using asyncio
async def main():
    # input = {"prompt": "a cartoon of a IRONMAN fighting with HULK, wall painting"}
    # client = Client(api_token="<API_TOKEN_HERE>")
    # result = await client.run_async(
    #     version="001bb81139b01780380407b4106ac681df46108e002eafbeb9ccb2d8faca42e1",
    #     input=input,
    # )
    # print("Result:", result)
    a = [
        {
            "pytorch_version": "v2.2.2",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.2.2 torchvision==0.17.2 torchaudio==2.2.2 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.2.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.2.1 torchvision==0.17.1 torchaudio==2.2.1 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.2.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.2.0 torchvision==0.17.0 torchaudio==2.2.0 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.1.2",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.1.2 torchvision==0.16.2 torchaudio==2.1.2 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.1.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.1.1 torchvision==0.16.1 torchaudio==2.1.1 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.1.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "12.1",
                            "install_cmd": "pip install torch==2.1.0 torchvision==0.16.0 torchaudio==2.1.0 --index-url https://download.pytorch.org/whl/cu121",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.0.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.8",
                            "install_cmd": "pip install torch==2.0.1 torchvision==0.15.2 torchaudio==2.0.2 --index-url https://download.pytorch.org/whl/cu118",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v2.0.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.8",
                            "install_cmd": "pip install torch==2.0.0 torchvision==0.15.1 torchaudio==2.0.1 --index-url https://download.pytorch.org/whl/cu118",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.13.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.7",
                            "install_cmd": "pip install torch==1.13.1+cu117 torchvision==0.14.1+cu117 torchaudio==0.13.1 --extra-index-url https://download.pytorch.org/whl/cu117",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.13.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.7",
                            "install_cmd": "pip install torch==1.13.0+cu117 torchvision==0.14.0+cu117 torchaudio==0.13.0 --extra-index-url https://download.pytorch.org/whl/cu117",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.12.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.6",
                            "install_cmd": "pip install torch==1.12.1+cu116 torchvision==0.13.1+cu116 torchaudio==0.12.1 --extra-index-url https://download.pytorch.org/whl/cu116",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.12.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.6",
                            "install_cmd": "pip install torch==1.12.0+cu116 torchvision==0.13.0+cu116 torchaudio==0.12.0 --extra-index-url https://download.pytorch.org/whl/cu116",
                        }
                    ]
                }
            },
        },
    ] + [
        {
            "pytorch_version": "v1.11.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.5",
                            "install_cmd": "pip install torch==1.11.0+cu115 torchvision==0.12.0+cu115 torchaudio==0.11.0 --extra-index-url https://download.pytorch.org/whl/cu115",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.10.2",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.3",
                            "install_cmd": "pip install torch==1.10.2+cu113 torchvision==0.11.3+cu113 torchaudio==0.10.2 --extra-index-url https://download.pytorch.org/whl/cu113",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.10.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.3",
                            "install_cmd": "pip install torch==1.10.1+cu113 torchvision==0.11.2+cu113 torchaudio==0.10.1 --extra-index-url https://download.pytorch.org/whl/cu113",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.10.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.3",
                            "install_cmd": "pip install torch==1.10.0+cu113 torchvision==0.11.1+cu113 torchaudio==0.10.0 --extra-index-url https://download.pytorch.org/whl/cu113",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.9.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.1",
                            "install_cmd": "pip install torch==1.9.1+cu111 torchvision==0.10.1+cu111 torchaudio==0.9.1 --extra-index-url https://download.pytorch.org/whl/cu111",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.9.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.1",
                            "install_cmd": "pip install torch==1.9.0+cu111 torchvision==0.10.0+cu111 torchaudio==0.9.0 --extra-index-url https://download.pytorch.org/whl/cu111",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.8.2",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.1",
                            "install_cmd": "pip install torch==1.8.2+cu111 torchvision==0.9.2+cu111 torchaudio==0.8.2 --extra-index-url https://download.pytorch.org/whl/cu111",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.8.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.1",
                            "install_cmd": "pip install torch==1.8.1+cu111 torchvision==0.9.1+cu111 torchaudio==0.8.1 --extra-index-url https://download.pytorch.org/whl/cu111",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.8.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.1",
                            "install_cmd": "pip install torch==1.8.0+cu111 torchvision==0.9.0+cu111 torchaudio==0.8.0 --extra-index-url https://download.pytorch.org/whl/cu111",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.7.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.0",
                            "install_cmd": "pip install torch==1.7.1+cu110 torchvision==0.8.2+cu110 torchaudio==0.7.2 --extra-index-url https://download.pytorch.org/whl/cu110",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.7.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "11.0",
                            "install_cmd": "pip install torch==1.7.0+cu110 torchvision==0.8.1+cu110 torchaudio==0.7.0 --extra-index-url https://download.pytorch.org/whl/cu110",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.6.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.2",
                            "install_cmd": "pip install torch==1.6.0+cu102 torchvision==0.7.0+cu102 torchaudio==0.6.0 --extra-index-url https://download.pytorch.org/whl/cu102",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.5.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.2",
                            "install_cmd": "pip install torch==1.5.1+cu102 torchvision==0.6.1+cu102 torchaudio==0.5.1 --extra-index-url https://download.pytorch.org/whl/cu102",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.5.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.2",
                            "install_cmd": "pip install torch==1.5.0+cu102 torchvision==0.6.0+cu102 torchaudio==0.5.0 --extra-index-url https://download.pytorch.org/whl/cu102",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.4.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.1",
                            "install_cmd": "pip install torch==1.4.0+cu101 torchvision==0.5.0+cu101 torchaudio==0.4.0 --extra-index-url https://download.pytorch.org/whl/cu101",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.3.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.1",
                            "install_cmd": "pip install torch==1.3.1+cu101 torchvision==0.4.2+cu101 --extra-index-url https://download.pytorch.org/whl/cu101",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.3.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.1",
                            "install_cmd": "pip install torch==1.3.0+cu101 torchvision==0.4.1+cu101 --extra-index-url https://download.pytorch.org/whl/cu101",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.2.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.0",
                            "install_cmd": "pip install torch==1.2.0+cu100 torchvision==0.4.0+cu100 --extra-index-url https://download.pytorch.org/whl/cu100",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.1.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.0",
                            "install_cmd": "pip install torch==1.1.0+cu100 torchvision==0.3.0+cu100 --extra-index-url https://download.pytorch.org/whl/cu100",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.0.1",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "10.0",
                            "install_cmd": "pip install torch==1.0.1+cu100 torchvision==0.2.2+cu100 --extra-index-url https://download.pytorch.org/whl/cu100",
                        }
                    ]
                }
            },
        },
        {
            "pytorch_version": "v1.0.0",
            "Wheel": {
                "linux": {
                    "cuda_version": [
                        {
                            "cuda": "9.0",
                            "install_cmd": "pip install torch==1.0.0 torchvision==0.2.1 --extra-index-url https://download.pytorch.org/whl/cu90",
                        }
                    ]
                }
            },
        },
    ]


# Run the main function using asyncio.run
if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from pebble import ThreadPool
import logging
from fastapi import HTTPException


class RunnerResult:
    def __init__(self, msg):
        self.msg = msg


class Runner:
    def __init__(self) -> None:
        self.futures = {}

    async def predict(self, id, func, request_body):
        if id not in self.futures:
            with ThreadPool() as pool:
                future = pool.schedule(
                    func,
                    kwargs=request_body["input"],
                )
                self.futures[id] = future
                while not future.done():
                    logging.info("sleeping")
                    await asyncio.sleep(0.1)
                del self.futures[id]
                return future.result()
        else:
            return RunnerResult(
                "A prediction is already running with id -> {}".format(id)
            )

    def cancel(self, id) -> bool:
        if id not in self.futures:
            raise HTTPException(status_code=404, detail="id not found")
        result = self.futures[id]
        result.cancel()
        del self.futures[id]
        logging.debug("Cancel result {}".format(result))
        return result

    def clean_up(self):
        for id, future in self.futures.items():
            future.cancel()



from abc import ABC, abstractmethod
from typing import Any


class BasePredictor(ABC):

    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def predict(self, **kwargs: Any) -> Any:
        """
        Run a single prediction on the model
        """
        pass
